# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

import galsim
from galsim_test_helpers import *


@timer
def test_hlr():
    """Test the calculateHLR method.
    """
    # Compare the calculation for a simple Gaussian.
    g1 = galsim.Gaussian(sigma=5, flux=1.7)

    print('g1 native hlr = ',g1.half_light_radius)
    print('g1.calculateHLR = ',g1.calculateHLR())
    print('nyquist scale = ',g1.nyquist_scale)
    # These should be exactly equal.
    np.testing.assert_equal(g1.half_light_radius, g1.calculateHLR(),
                            err_msg="Gaussian.calculateHLR() returned wrong value.")

    # Check for a convolution of two Gaussians.  Should be equivalent, but now will need to
    # do the calculation.
    g2 = galsim.Convolve(galsim.Gaussian(sigma=3, flux=1.3), galsim.Gaussian(sigma=4, flux=23))
    test_hlr = g2.calculateHLR()
    print('g2.calculateHLR = ',test_hlr)
    print('ratio - 1 = ',test_hlr/g1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/g1.half_light_radius, 1.0, decimal=1,
                                   err_msg="Gaussian.calculateHLR() is not accurate.")

    # The default scale is only accurate to around 1 dp.  Using scale = 0.1 is accurate to 3 dp.
    # Note: Nyquist scale is about 4.23 for this profile.
    test_hlr = g2.calculateHLR(scale=0.1)
    print('g2.calculateHLR(scale=0.1) = ',test_hlr)
    print('ratio - 1 = ',test_hlr/g1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/g1.half_light_radius, 1.0, decimal=3,
                                   err_msg="Gaussian.calculateHLR(scale=0.1) is not accurate.")

    # Finally, we don't expect this to be accurate, but make sure the code can handle having
    # more than half the flux in the central pixel.
    test_hlr = g2.calculateHLR(scale=15)
    print('g2.calculateHLR(scale=15) = ',test_hlr)
    print('ratio - 1 = ',test_hlr/g1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/g1.half_light_radius/10, 0.1, decimal=1,
                                   err_msg="Gaussian.calculateHLR(scale=15) is not accurate.")

    # Next, use an Exponential profile
    e1 = galsim.Exponential(scale_radius=5, flux=1.7)

    print('e1 native hlr = ',e1.half_light_radius)
    print('e1.calculateHLR = ',e1.calculateHLR())
    print('nyquist scale = ',e1.nyquist_scale)
    # These should be exactly equal.
    np.testing.assert_equal(e1.half_light_radius, e1.calculateHLR(),
                            err_msg="Exponential.calculateHLR() returned wrong value.")

    # Check for a convolution with a delta function.  Should be equivalent, but now will need to
    # do the calculation.
    e2 = galsim.Convolve(galsim.Exponential(scale_radius=5, flux=1.3),
                         galsim.Gaussian(sigma=1.e-4, flux=23))
    test_hlr = e2.calculateHLR()
    print('e2.calculateHLR = ',test_hlr)
    print('ratio - 1 = ',test_hlr/e1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/e1.half_light_radius, 1.0, decimal=1,
                                   err_msg="Exponential.calculateHLR() is not accurate.")

    # The default scale is only accurate to around 1 dp.  Using scale = 0.1 is accurate to 3 dp.
    # Note: Nyquist scale is about 1.57 for this profile.
    # We can also decrease the size, which should still be accurate, but maybe a little faster.
    # Go a bit more that 2*hlr in units of the pixels.
    size = int(2.2 * e1.half_light_radius / 0.1)
    test_hlr = e2.calculateHLR(scale=0.1, size=size)
    print('e2.calculateHLR(scale=0.1) = ',test_hlr)
    print('ratio - 1 = ',test_hlr/e1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/e1.half_light_radius, 1.0, decimal=3,
                                   err_msg="Exponential.calculateHLR(scale=0.1) is not accurate.")

    # Check that it works if the centroid is not at the origin
    e3 = e2.shift(2,3)
    test_hlr = e3.calculateHLR(scale=0.1)
    print('e3.calculateHLR(scale=0.1) = ',test_hlr)
    print('ratio - 1 = ',test_hlr/e1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/e1.half_light_radius, 1.0, decimal=3,
                                   err_msg="shifted Exponential HLR is not accurate.")

    # Can set a centroid manually.  This should be equivalent to the default.
    print('e3.centroid = ',e3.centroid)
    test_hlr = e3.calculateHLR(scale=0.1, centroid=e3.centroid)
    np.testing.assert_almost_equal(test_hlr/e1.half_light_radius, 1.0, decimal=3,
                                   err_msg="shifted HLR with explicit centroid is not accurate.")

    # The calculateHLR method can also return other radii like r90, rather than r50 using the
    # parameter flux_fraction.  This is also analytic for Exponential
    r90 = 3.889720170 * e1.scale_radius
    test_r90 = e2.calculateHLR(scale=0.1, flux_frac=0.9)
    print('r90 = ',r90)
    print('e2.calculateHLR(scale=0.1, flux_frac=0.9) = ',test_r90)
    print('ratio - 1 = ',test_r90/r90-1)
    np.testing.assert_almost_equal(test_r90/r90, 1.0, decimal=3,
                                   err_msg="Exponential r90 calculation is not accurate.")

    # Check the image version.
    im = e1.drawImage(scale=0.1, nx=2048, ny=2048)  # Needs to be large to get enough flux for the
                                                    # image to get it to 3 digits of accuracy.
    test_hlr = im.calculateHLR()
    print('im.calculateHLR() = ',test_hlr)
    print('ratio - 1 = ',test_hlr/e1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/e1.half_light_radius, 1.0, decimal=3,
                                   err_msg="image.calculateHLR is not accurate.")

    # Check that a non-square image works correctly.  Also, not centered anywhere in particular.
    bounds = galsim.BoundsI(-1234, -1234+2048, 8234, 8234+2099)
    offset = galsim.PositionD(29,1)
    im = e1.drawImage(scale=0.1, bounds=bounds, offset=offset)
    test_hlr = im.calculateHLR(center=im.true_center+offset)
    print('im.calculateHLR() = ',test_hlr)
    print('ratio - 1 = ',test_hlr/e1.half_light_radius-1)
    np.testing.assert_almost_equal(test_hlr/e1.half_light_radius, 1.0, decimal=3,
                                   err_msg="non-square image.calculateHLR is not accurate.")


@timer
def test_sigma():
    """Test the calculateMomentRadius method.
    """
    # Compare the calculation for a simple Gaussian.
    g1 = galsim.Gaussian(sigma=5, flux=1.7)

    print('g1 native sigma = ',g1.sigma)
    print('g1.calculateMomentRadius = ',g1.calculateMomentRadius())
    # These should be exactly equal.
    np.testing.assert_equal(
            g1.sigma, g1.calculateMomentRadius(),
            err_msg="Gaussian.calculateMomentRadius() returned wrong value.")
    np.testing.assert_equal(
            g1.sigma, g1.calculateMomentRadius(rtype='trace'),
            err_msg="Gaussian.calculateMomentRadius(trace) returned wrong value.")
    np.testing.assert_equal(
            g1.sigma, g1.calculateMomentRadius(rtype='det'),
            err_msg="Gaussian.calculateMomentRadius(det) returned wrong value.")
    np.testing.assert_equal(
            (g1.sigma, g1.sigma), g1.calculateMomentRadius(rtype='both'),
            err_msg="Gaussian.calculateMomentRadius(both) returned wrong value.")
    with assert_raises(galsim.GalSimValueError):
        g1.calculateMomentRadius(rtype='invalid')

    # Check for a convolution of two Gaussians.  Should be equivalent, but now will need to
    # do the calculation.
    g2 = galsim.Convolve(galsim.Gaussian(sigma=3, flux=1.3), galsim.Gaussian(sigma=4, flux=23))
    test_sigma = g2.calculateMomentRadius()
    print('g2.calculateMomentRadius = ',test_sigma)
    print('ratio - 1 = ',test_sigma/g1.sigma-1)
    np.testing.assert_almost_equal(
            test_sigma/g1.sigma, 1.0, decimal=1,
            err_msg="Gaussian.calculateMomentRadius() is not accurate.")

    # The default scale and size is only accurate to around 1 dp.  Using scale = 0.1 is accurate
    # to 4 dp.
    test_sigma = g2.calculateMomentRadius(scale=0.1)
    print('g2.calculateMomentRadius(scale=0.1) = ',test_sigma)
    print('ratio - 1 = ',test_sigma/g1.sigma-1)
    np.testing.assert_almost_equal(
            test_sigma/g1.sigma, 1.0, decimal=4,
            err_msg="Gaussian.calculateMomentRadius(scale=0.1) is not accurate.")

    # In this case, the different calculations are eqivalent:
    np.testing.assert_almost_equal(
            test_sigma, g2.calculateMomentRadius(scale=0.1, rtype='trace'),
            err_msg="Gaussian.calculateMomentRadius(trace) is not accurate.")
    np.testing.assert_almost_equal(
            test_sigma, g2.calculateMomentRadius(scale=0.1, rtype='det'),
            err_msg="Gaussian.calculateMomentRadius(trace) is not accurate.")
    np.testing.assert_almost_equal(
            (test_sigma, test_sigma), g2.calculateMomentRadius(scale=0.1, rtype='both'),
            err_msg="Gaussian.calculateMomentRadius(trace) is not accurate.")

    # However, when we shear it, the default (det) measure stays equal to the original sigma, but
    # the trace measure increases by a factor of (1-e^2)^0.25
    g3 = g2.shear(e1=0.4, e2=0.3)
    esq = 0.4**2 + 0.3**2
    sheared_sigma = g3.calculateMomentRadius(scale=0.1)
    print('g3.calculateMomentRadius(scale=0.1) = ',sheared_sigma)
    print('ratio - 1 = ',sheared_sigma/g1.sigma-1)
    sheared_sigma2 = g3.calculateMomentRadius(scale=0.1, rtype='trace')
    print('g3.calculateMomentRadius(scale=0.1,trace) = ',sheared_sigma2)
    print('ratio = ',sheared_sigma2 / g1.sigma)
    print('(1-e^2)^-0.25 = ',(1-esq)**-0.25)
    print('ratio - 1 = ',sheared_sigma2/(g1.sigma*(1.-esq)**-0.25)-1)
    np.testing.assert_almost_equal(
            sheared_sigma/g1.sigma, 1.0, decimal=4,
            err_msg="sheared Gaussian.calculateMomentRadius(scale=0.1) is not accurate.")
    np.testing.assert_almost_equal(
            sheared_sigma2/(g1.sigma*(1.-esq)**-0.25), 1.0, decimal=4,
            err_msg="sheared Gaussian.calculateMomentRadius(scale=0.1,trace) is not accurate.")


    # Next, use an Exponential profile
    e1 = galsim.Exponential(scale_radius=5, flux=1.7)

    # The true "sigma" for this is analytic, but not an attribute.
    e1_sigma = np.sqrt(3.0) * e1.scale_radius
    print('true e1 sigma = sqrt(3) * e1.scale_radius = ',e1_sigma)

    # Test with the default scale and size.
    test_sigma = e1.calculateMomentRadius()
    print('e1.calculateMomentRadius = ',test_sigma)
    print('ratio - 1 = ',test_sigma/e1_sigma-1)
    np.testing.assert_almost_equal(
            test_sigma/e1_sigma, 1.0, decimal=1,
            err_msg="Exponential.calculateMomentRadius() is not accurate.")

    # The default scale and size is only accurate to around 1 dp.  This time we have to both
    # decrease the scale and also increase the size to get 4 dp of precision.
    test_sigma = e1.calculateMomentRadius(scale=0.1, size=2000)
    print('e1.calculateMomentRadius(scale=0.1) = ',test_sigma)
    print('ratio - 1 = ',test_sigma/e1_sigma-1)
    np.testing.assert_almost_equal(
            test_sigma/e1_sigma, 1.0, decimal=4,
            err_msg="Exponential.calculateMomentRadius(scale=0.1) is not accurate.")

    # Check that it works if the centroid is not at the origin
    e3 = e1.shift(2,3)
    test_sigma = e3.calculateMomentRadius(scale=0.1, size=2000)
    print('e1.calculateMomentRadius(scale=0.1) = ',test_sigma)
    print('ratio - 1 = ',test_sigma/e1_sigma-1)
    np.testing.assert_almost_equal(
            test_sigma/e1_sigma, 1.0, decimal=4,
            err_msg="shifted Exponential MomentRadius is not accurate.")

    # Can set a centroid manually.  This should be equivalent to the default.
    print('e3.centroid = ',e3.centroid)
    test_sigma = e3.calculateMomentRadius(scale=0.1, size=2000, centroid=e3.centroid)
    np.testing.assert_almost_equal(
            test_sigma/e1_sigma, 1.0, decimal=4,
            err_msg="shifted MomentRadius with explicit centroid is not accurate.")

    # Check the image version.
    size = 2000
    im = e1.drawImage(scale=0.1, nx=size, ny=size)
    test_sigma = im.calculateMomentRadius()
    print('im.calculateMomentRadius() = ',test_sigma)
    print('ratio - 1 = ',test_sigma/e1_sigma-1)
    np.testing.assert_almost_equal(
            test_sigma/e1_sigma, 1.0, decimal=4,
            err_msg="image.calculateMomentRadius is not accurate.")
    with assert_raises(galsim.GalSimValueError):
        im.calculateMomentRadius(rtype='invalid')

    # Check that a non-square image works correctly.  Also, not centered anywhere in particular.
    bounds = galsim.BoundsI(-1234, -1234+size*2, 8234, 8234+size)
    offset = galsim.PositionD(29,1)
    im = e1.drawImage(scale=0.1, bounds=bounds, offset=offset)
    test_hlr = im.calculateMomentRadius(center=im.true_center+offset)
    print('im.calculateMomentRadius() = ',test_sigma)
    print('ratio - 1 = ',test_sigma/e1_sigma-1)
    np.testing.assert_almost_equal(
            test_sigma/e1_sigma, 1.0, decimal=4,
            err_msg="non-square image.calculateMomentRadius is not accurate.")


@timer
def test_fwhm():
    """Test the calculateFWHM method.
    """
    # Compare the calculation for a simple Gaussian.
    g1 = galsim.Gaussian(sigma=5, flux=1.7)

    print('g1 native fwhm = ',g1.fwhm)
    print('g1.calculateFWHM = ',g1.calculateFWHM())
    # These should be exactly equal.
    np.testing.assert_equal(g1.fwhm, g1.calculateFWHM(),
                            err_msg="Gaussian.calculateFWHM() returned wrong value.")

    # Check for a convolution of two Gaussians.  Should be equivalent, but now will need to
    # do the calculation.
    g2 = galsim.Convolve(galsim.Gaussian(sigma=3, flux=1.3), galsim.Gaussian(sigma=4, flux=23))
    test_fwhm = g2.calculateFWHM()
    print('g2.calculateFWHM = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/g1.fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/g1.fwhm, 1.0, decimal=3,
                                   err_msg="Gaussian.calculateFWHM() is not accurate.")

    # The default scale already accurate to around 3 dp.  Using scale = 0.1 is accurate to 8 dp.
    test_fwhm = g2.calculateFWHM(scale=0.1)
    print('g2.calculateFWHM(scale=0.1) = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/g1.fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/g1.fwhm, 1.0, decimal=8,
                                   err_msg="Gaussian.calculateFWHM(scale=0.1) is not accurate.")

    # Finally, we don't expect this to be accurate, but make sure the code can handle having
    # only the central pixel higher than half-maximum.
    test_fwhm = g2.calculateFWHM(scale=20)
    print('g2.calculateFWHM(scale=20) = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/g1.fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/g1.fwhm/10, 0.1, decimal=1,
                                   err_msg="Gaussian.calculateFWHM(scale=20) is not accurate.")

    # Next, use an Exponential profile
    e1 = galsim.Exponential(scale_radius=5, flux=1.7)

    # The true fwhm for this is analytic, but not an attribute.
    e1_fwhm = 2. * np.log(2.0) * e1.scale_radius
    print('true e1 fwhm = ',e1_fwhm)

    # Test with the default scale and size.
    test_fwhm = e1.calculateFWHM()
    print('e1.calculateFWHM = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/e1_fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/e1_fwhm, 1.0, decimal=3,
                                   err_msg="Exponential.calculateFWHM() is not accurate.")

    # The default scale already accurate to around 3 dp.  Using scale = 0.1 is accurate to 7 dp.
    # We can also decrease the size, which should still be accurate, but maybe a little faster.
    # Go a bit more that fwhm in units of the pixels.
    size = int(1.2 * e1_fwhm / 0.1)
    test_fwhm = e1.calculateFWHM(scale=0.1, size=size)
    print('e1.calculateFWHM(scale=0.1) = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/e1_fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/e1_fwhm, 1.0, decimal=7,
                                   err_msg="Exponential.calculateFWHM(scale=0.1) is not accurate.")

    # Check that it works if the centroid is not at the origin
    e3 = e1.shift(2,3)
    test_fwhm = e3.calculateFWHM(scale=0.1)
    print('e3.calculateFWHM(scale=0.1) = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/e1_fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/e1_fwhm, 1.0, decimal=6,
                                   err_msg="shifted Exponential FWHM is not accurate.")

    # Can set a centroid manually.  This should be equivalent to the default.
    print('e3.centroid = ',e3.centroid)
    test_fwhm = e3.calculateFWHM(scale=0.1, centroid=e3.centroid)
    np.testing.assert_almost_equal(test_fwhm/e1_fwhm, 1.0, decimal=6,
                                   err_msg="shifted FWHM with explicit centroid is not accurate.")

    # Check the image version.
    im = e1.drawImage(scale=0.1, method='sb')
    test_fwhm = im.calculateFWHM(Imax=e1.xValue(0,0))
    print('im.calculateFWHM() = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/e1_fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/e1_fwhm, 1.0, decimal=6,
                                   err_msg="image.calculateFWHM is not accurate.")

    # Check that a non-square image works correctly.  Also, not centered anywhere in particular.
    bounds = galsim.BoundsI(-1234, -1234+size*2, 8234, 8234+size)
    offset = galsim.PositionD(29,1)
    im = e1.drawImage(scale=0.1, bounds=bounds, offset=offset, method='sb')
    test_fwhm = im.calculateFWHM(Imax=e1.xValue(0,0), center=im.true_center+offset)
    print('im.calculateFWHM() = ',test_fwhm)
    print('ratio - 1 = ',test_fwhm/e1_fwhm-1)
    np.testing.assert_almost_equal(test_fwhm/e1_fwhm, 1.0, decimal=6,
                                   err_msg="non-square image.calculateFWHM is not accurate.")


def test_fwhm_deterministic_fwhm():
    # This test was proposed by Wiliam Lucas in issue #1336.
    # Numpy changed some details about how argsort works, so different numpy versions had
    # given different answers to the following calculation.

    gal = galsim.Gaussian(flux=1.e5, sigma=5.)
    rng = galsim.BaseDeviate(1234)
    img = gal.drawImage(method='phot', rng=rng, nx=50, ny=50)
    fwhm = img.calculateFWHM()
    # This value used to be different on systems with different numpy versions.
    # Note: this only affects situations where the center is a pixel center or pixel corner
    # (or some other very special locations) such that the rsq values include repeated values.
    # Then argsort could return a different data index for what pixel was the first one past
    # the half max value.  For symmetric profiles, even this probalby wasn't a problem, since all
    # data values at that radius have the same value.  But for photon shooting, there is photon
    # noise, which means the values are not identical, leading to indeterminacy in the FWHM value.
    print(fwhm)

    # The code was changed in GalSim version 2.7 to be insensitive to the details of what argsort
    # does with repeated values.  Specifically, we now take the mean of all data values with the
    # same rsq value.
    # Thus, this value should be consistent across any algorithmic changes numpy might make to
    # how argsort works in detail.
    assert np.isclose(fwhm, 13.715780373024351)

if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os
import numpy as np

import galsim
from galsim_test_helpers import *


path, filename = os.path.split(__file__)
datapath = os.path.abspath(os.path.join(path, "../examples/data/"))


@timer
def test_cosmos_basic():
    """Check some basic functionality of the COSMOSCatalog class."""
    # Note, there's not much here yet.   Could try to think of other tests that are more
    # interesting.

    # Initialize a COSMOSCatalog with all defaults.
    cat = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                               dir=datapath)

    # Check GalaxySample equivalent
    cat1 = galsim.GalaxySample('real_galaxy_catalog_23.5_example.fits', datapath,
                               cut_ratio=0.2, sn_limit=20.)
    assert cat1.nobjects == cat1.getNObjects() == cat.nobjects == cat.getNObjects()
    assert cat1 == cat  # These (intentionally) test as equal even though different classes.

    # Initialize one that doesn't exclude failures.  It should be >= the previous one in length.
    cat2 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                               dir=datapath, exclusion_level='none')
    assert cat2.nobjects >= cat.nobjects
    assert len(cat2) == cat2.nobjects == cat2.getNTot() == 100
    assert len(cat) == cat.nobjects < cat.getNTot()

    # Check other exclusion levels:
    cat3 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                dir=datapath, exclusion_level='bad_stamp')
    assert len(cat3) == 97
    cat4 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                dir=datapath, exclusion_level='bad_fits')
    assert len(cat4) == 100  # no bad fits in the example file as it happens.
    cat5 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                dir=datapath, exclusion_level='marginal')
    assert len(cat5) == 96   # this is actually the default, so == cat
    assert cat == cat5
    cat6 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                dir=datapath, exclusion_level='marginal',
                                max_hlr=2, max_flux=2000)
    assert len(cat6) == 93
    cat7 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                dir=datapath, exclusion_level='marginal',
                                min_hlr=0.2, min_flux=10)
    assert len(cat7) == 91

    # Check the 25.2 exclusions.  We don't have a 25.2 catalog available in Travis runs, so
    # mock up the example catalog as though it were 25.2
    # Also check the min/max hlr and flux options.
    cat2.use_sample = '25.2'
    hlr = cat2.param_cat['hlr'][:,0]
    flux = cat2.param_cat['flux'][:,0]
    print("Full range of hlr = ", np.min(hlr), np.max(hlr))
    print("Full range of flux = ", np.min(flux), np.max(flux))
    cat2._apply_exclusion('marginal', min_hlr=0.2, max_hlr=2, min_flux=50, max_flux=5000)
    print("Size of catalog with hlr and flux exclusions == ",len(cat2))
    assert len(cat2) == 47

    # Repeat with a fake 25.2 catalog (just symlinks to the above 23.5 catalog).
    fake_25_2_cat = galsim.COSMOSCatalog(file_name='fake_25.2.fits', dir='input',
                                         exclusion_level='marginal', min_hlr=0.2, max_hlr=2,
                                         min_flux=50, max_flux=5000)
    assert len(fake_25_2_cat) == 47

    # Check for reasonable exceptions when initializing.
    # Can't find data (wrong directory).
    with assert_raises(OSError):
        galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits')

    # Try making galaxies
    gal_real = cat.makeGalaxy(index=0,gal_type='real',chromatic=False)
    assert isinstance(gal_real, galsim.RealGalaxy)

    gal_param = cat.makeGalaxy(index=10,gal_type='parametric',chromatic=True)
    assert isinstance(gal_param, galsim.ChromaticObject)

    # Second time through, don't make the bandpass.
    bp = cat._bandpass
    sed = cat._sed
    assert bp is not None
    gal_param2 = cat.makeGalaxy(index=13, gal_type='parametric', chromatic=True)
    assert isinstance(gal_param2, galsim.ChromaticObject)
    assert gal_param != gal_param2
    assert cat._bandpass is bp   # Not just ==.  "is" means the same object.
    assert cat._sed is sed

    # So far, we've made a bulge+disk and a disky Sersic.
    # Do two more to run through two more paths in the code.
    gal_param3 = cat.makeGalaxy(index=50, gal_type='parametric', chromatic=True)
    gal_param4 = cat.makeGalaxy(index=67, gal_type='parametric', chromatic=True)

    gal_real_list = cat.makeGalaxy(index=[3,6],gal_type='real',chromatic=False)
    for gal_real in gal_real_list:
        assert isinstance(gal_real, galsim.RealGalaxy)

    gal_param_list = cat.makeGalaxy(index=[4,7],gal_type='parametric',chromatic=False)
    for gal_param in gal_param_list:
        assert isinstance(gal_param, galsim.GSObject)

    # Check for parametric catalog
    # Can give either the regular name or the _fits name.
    cat_param = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                     dir=datapath, use_real=False)
    cat_param2 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example_fits.fits',
                                      dir=datapath, use_real=False)
    assert cat_param2 == cat_param

    # Try making galaxies
    gal = cat_param.makeGalaxy(index=1)
    assert isinstance(gal, galsim.GSObject)

    gal_list = cat_param.makeGalaxy(index=[2,3])
    for gal in gal_list:
        assert isinstance(gal, galsim.GSObject)

    # Check sersic_prec option.
    sersic0 = cat_param.makeGalaxy(index=59, sersic_prec=0)
    np.testing.assert_almost_equal(sersic0.original.n, 1.14494567108)
    sersic1 = cat_param.makeGalaxy(index=59, sersic_prec=0.05)  # The default.
    np.testing.assert_almost_equal(sersic1.original.n, 1.15)
    sersic2 = cat_param.makeGalaxy(index=59, sersic_prec=0.1)
    np.testing.assert_almost_equal(sersic2.original.n, 1.1)
    sersic3 = cat_param.makeGalaxy(index=59, sersic_prec=0.5)
    np.testing.assert_almost_equal(sersic3.original.n, 1.0)

    assert_raises(TypeError, galsim.COSMOSCatalog, 'real_galaxy_catalog_23.5_example.fits',
                  dir=datapath, sample='23.5')
    assert_raises(ValueError, galsim.COSMOSCatalog, sample='invalid')

    assert_raises(ValueError, cat_param.makeGalaxy, gal_type='real')
    assert_raises(ValueError, cat_param.makeGalaxy, gal_type='invalid')
    assert_raises(ValueError, cat.makeGalaxy, gal_type='invalid')
    assert_raises(TypeError, cat_param.makeGalaxy, rng='invalid')

    assert_raises(NotImplementedError, cat.makeGalaxy, gal_type='real', chromatic=True)

@timer
def test_cosmos_fluxnorm():
    """Check for flux normalization properties of COSMOSCatalog class."""
    # Check that if we make a RealGalaxy catalog, and a COSMOSCatalog, and draw the real object, the
    # fluxes should match very well.  These correspond to 1s exposures.
    test_ind = 54
    rand_seed = 12345
    cat = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                               dir=datapath, exclusion_level='none')
    rgc = galsim.RealGalaxyCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                   dir=datapath)
    final_psf = galsim.Airy(diam=1.2, lam=800.) # PSF twice as big as HST in F814W.
    gal1 = cat.makeGalaxy(test_ind, gal_type='real', rng=galsim.BaseDeviate(rand_seed))
    gal2 = galsim.RealGalaxy(rgc, index=test_ind, rng=galsim.BaseDeviate(rand_seed))
    gal1 = galsim.Convolve(gal1, final_psf)
    gal2 = galsim.Convolve(gal2, final_psf)
    im1 = gal1.drawImage(scale=0.05)
    im2 = gal2.drawImage(scale=0.05)

    # Then check that if we draw a parametric representation that is achromatic, that the flux
    # matches reasonably well (won't be exact because model isn't perfect).
    gal1_param = cat.makeGalaxy(test_ind, gal_type='parametric', chromatic=False)
    gal1_param_final = galsim.Convolve(gal1_param, final_psf)
    im1_param = gal1_param_final.drawImage(scale=0.05)

    # Then check the same for a chromatic parametric representation that is drawn into the same
    # band.
    bp_file = os.path.join(galsim.meta_data.share_dir, 'bandpasses', 'ACS_wfc_F814W.dat')
    bandpass = galsim.Bandpass(bp_file, wave_type='nm').withZeropoint(25.94)#34.19)
    gal1_chrom = cat.makeGalaxy(test_ind, gal_type='parametric', chromatic=True)
    gal1_chrom = galsim.Convolve(gal1_chrom, final_psf)
    im1_chrom = gal1_chrom.drawImage(bandpass, scale=0.05)

    ref_val = [im1.array.sum(), im1.array.sum(), im1.array.sum()]
    test_val = [im2.array.sum(), im1_param.array.sum(), im1_chrom.array.sum()]
    np.testing.assert_allclose(ref_val, test_val, rtol=0.1,
                               err_msg='Flux normalization problem in COSMOS galaxies')

    # Finally, check that the original COSMOS info is stored properly after transformations, for
    # both Sersic galaxies (like galaxy 0 in the catalog) and the one that is gal1_param above.
    gal0_param = cat.makeGalaxy(0, gal_type='parametric', chromatic=False)
    assert hasattr(gal0_param.shear(g1=0.05).original, 'index'), \
        'Sersic galaxy does not retain index information after transformation'
    assert hasattr(gal1_param.shear(g1=0.05).original, 'index'), \
        'Bulge+disk galaxy does not retain index information after transformation'

    assert_raises(ValueError, galsim.COSMOSCatalog, 'real_galaxy_catalog_23.5_example.fits',
                  dir=datapath, exclusion_level='invalid')

    # Check scaling with area and exptime
    print('hst area = ',galsim.COSMOSCatalog.hst_eff_area)
    assert np.isclose(galsim.COSMOSCatalog.hst_eff_area,
                      np.pi * 2.4**2 / 4 * (1-0.33**2) * 100**2)
    cat2 = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                dir=datapath, exclusion_level='none', area=3456, exptime=2.3)
    flux_scaling = 2.3 * (3456/galsim.COSMOSCatalog.hst_eff_area)
    gal1 = cat.makeGalaxy(test_ind, gal_type='parametric')
    gal2 = cat2.makeGalaxy(test_ind, gal_type='parametric')
    assert np.isclose(gal2.flux, gal1.flux * flux_scaling)

@timer
def test_cosmos_random():
    """Check the random object functionality of the COSMOS catalog."""
    # For most of this test, we'll use the selectRandomIndex() routine, which does not try to
    # construct the GSObjects.  This makes the test go fast.  However, we will at the end have a
    # test to ensure that calling makeGalaxy() while requesting a random object has the same
    # behavior as using selectRandomIndex() in limited cases.

    # Initialize the catalog.  The first will have weights, while the second will not (since they
    # are in the RealGalaxyCatalog).
    cat = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                               dir=datapath)
    cat_param = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                     dir=datapath, use_real=False)
    assert cat_param.real_cat is None
    assert cat.real_cat is not None

    # Check for exception handling if bad inputs given for the random functionality.
    assert_raises(ValueError, cat.selectRandomIndex, 0)
    assert_raises(ValueError, cat.selectRandomIndex, 10.7)
    assert_raises(TypeError, cat.selectRandomIndex, 10, rng=3)

    # Check that random objects give the right <weight> without/with weighting.
    wt = cat.real_cat.weight[cat.orig_index]
    wt /= np.max(wt)
    avg_weight_val = np.sum(wt)/len(wt)
    wavg_weight_val = np.sum(wt**2)/np.sum(wt)
    with assert_raises(AssertionError):
        np.testing.assert_almost_equal(avg_weight_val, wavg_weight_val, 3)
    # Make sure we use enough objects that the mean weights converge properly.
    randind_wt = cat.selectRandomIndex(30000, rng=galsim.BaseDeviate(1234))
    wtrand = cat.real_cat.weight[cat.orig_index[randind_wt]] / \
        np.max(cat.real_cat.weight[cat.orig_index])
    # The average value of wtrand should be wavgw_weight_val in this case, since we used the weights
    # to probabilistically select galaxies.
    np.testing.assert_almost_equal(np.mean(wtrand), wavg_weight_val,3,
                                   err_msg='Average weight for random sample is wrong')

    # The param-only catalog doesn't have weights, so it does unweighted selection, which emits a
    # warning.  We know about this and want to ignore it here.
    with assert_warns(galsim.GalSimWarning):
        randind = cat_param.selectRandomIndex(30000, rng=galsim.BaseDeviate(1234))
    wtrand = cat.real_cat.weight[cat.orig_index[randind]] / \
        np.max(cat.real_cat.weight[cat.orig_index])
    # The average value of wtrand should be avg_weight_val, since we did not do a weighted
    # selection.
    np.testing.assert_almost_equal(np.mean(wtrand), avg_weight_val,3,
                                   err_msg='Average weight for random sample is wrong')

    # Check for consistency of randoms with same random seed.  Do this both for the weighted and the
    # unweighted calculation.
    # Check for inconsistency of randoms with different random seed, or same seed but without/with
    # weighting.
    rng1 = galsim.BaseDeviate(1234)
    rng2 = galsim.BaseDeviate(1234)
    ind1 = cat.selectRandomIndex(10, rng=rng1)
    ind2 = cat.selectRandomIndex(10, rng=rng2)
    np.testing.assert_array_equal(ind1,ind2,
                                  err_msg='Different random indices selected with same seed!')
    with assert_warns(galsim.GalSimWarning):
        ind1p = cat_param.selectRandomIndex(10, rng=rng1)
    with assert_warns(galsim.GalSimWarning):
        ind2p = cat_param.selectRandomIndex(10, rng=rng2)
    np.testing.assert_array_equal(ind1p,ind2p,
                                  err_msg='Different random indices selected with same seed!')
    rng3 = galsim.BaseDeviate(5678)
    ind3 = cat.selectRandomIndex(10, rng=rng3)
    ind3p = cat.selectRandomIndex(10) # initialize RNG based on time
    assert_raises(AssertionError, np.testing.assert_array_equal, ind1, ind1p)
    assert_raises(AssertionError, np.testing.assert_array_equal, ind1, ind3)
    assert_raises(AssertionError, np.testing.assert_array_equal, ind1p, ind3p)

    # Finally, make sure that directly calling selectRandomIndex() gives the same random ones as
    # makeGalaxy().  We'll do one real object because they are slower, and multiple parametric (just
    # to make sure that the multi-object selection works consistently).
    use_seed = 567
    obj = cat.makeGalaxy(rng=galsim.BaseDeviate(use_seed))
    ind = cat.selectRandomIndex(1, rng=galsim.BaseDeviate(use_seed))
    obj_2 = cat.makeGalaxy(ind)
    # Note: for real galaxies we cannot require that obj==obj_2, just that obj.index==obj_2.index.
    # That's because we want to make sure the same galaxy is being randomly selected, but we cannot
    # require that noise padding be the same, given the inconsistency in how the BaseDeviates are
    # used in the above cases.
    assert obj.index==obj_2.index,'makeGalaxy selects random objects inconsistently'

    n_random = 3
    objs = cat.makeGalaxy(rng=galsim.BaseDeviate(use_seed), gal_type='parametric',
                          n_random=n_random)
    inds = cat.selectRandomIndex(n_random, rng=galsim.BaseDeviate(use_seed))
    objs_2 = cat_param.makeGalaxy(inds)
    for i in range(n_random):
        # With parametric objects there is no noise padding, so we can require completely identical
        # GSObjects (not just equal indices).
        assert objs[i]==objs_2[i],'makeGalaxy selects random objects inconsistently'

    # Finally, check for consistency with random object selected from RealGalaxyCatalog.  For this
    # case, we need to make another COSMOSCatalog that does not flag the bad objects.
    use_seed=31415
    cat = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                               dir=datapath, exclusion_level='none')
    rgc = galsim.RealGalaxyCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                                   dir=datapath)
    ind_cc = cat.selectRandomIndex(1, rng=galsim.BaseDeviate(use_seed))
    foo = galsim.RealGalaxy(rgc, random=True, rng=galsim.BaseDeviate(use_seed))
    ind_rgc = foo.index
    assert ind_cc==ind_rgc,\
        'Different weighted random index selected from COSMOSCatalog and RealGalaxyCatalog'

    # Also check for the unweighted case.  Just remove that info from the catalogs and redo the
    # test.
    cat.real_cat = None
    rgc.weight = None
    with assert_warns(galsim.GalSimWarning):
        ind_cc = cat.selectRandomIndex(1, rng=galsim.BaseDeviate(use_seed))
    foo = galsim.RealGalaxy(rgc, random=True, rng=galsim.BaseDeviate(use_seed))
    ind_rgc = foo.index
    assert ind_cc==ind_rgc,\
        'Different unweighted random index selected from COSMOSCatalog and RealGalaxyCatalog'

    # Check that setting _n_rng_calls properly tracks the RNG calls for n_random=1 and >1.
    cat = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits',
                               dir=datapath, exclusion_level='none')
    test_seed = 123456
    ud = galsim.UniformDeviate(test_seed)
    obj, n_rng_calls = cat.selectRandomIndex(1, rng=ud, _n_rng_calls=True)
    ud2 = galsim.UniformDeviate(test_seed)
    ud2.discard(n_rng_calls)
    assert ud()==ud2(), '_n_rng_calls kwarg did not give proper tracking of RNG calls'
    ud = galsim.UniformDeviate(test_seed)
    obj, n_rng_calls = cat.selectRandomIndex(17, rng=ud, _n_rng_calls=True)
    ud2 = galsim.UniformDeviate(test_seed)
    ud2.discard(n_rng_calls)
    assert ud()==ud2(), '_n_rng_calls kwarg did not give proper tracking of RNG calls'

    # Invalid to both privide index and ask for random selection
    with assert_raises(galsim.GalSimIncompatibleValuesError):
        cat_param.makeGalaxy(index=(11,13,17), n_random=3)

@timer
def test_cosmos_deep():
    """Test the deep option of makeGalaxy"""

    cat = galsim.COSMOSCatalog(file_name='real_galaxy_catalog_23.5_example.fits', dir=datapath)

    # Pick a random galaxy
    # Turn off noise padding to make the comparisons more reliable.
    gal_shallow = cat.makeGalaxy(index=17, gal_type='real', noise_pad_size=0)
    print('gal_shallow = ',gal_shallow)
    shallow_flux = gal_shallow.flux
    shallow_hlr = gal_shallow.calculateHLR()
    print('flux = ',shallow_flux)
    print('hlr = ',shallow_hlr)

    gal_deep = cat.makeGalaxy(index=17, gal_type='real', deep=True, noise_pad_size=0)
    print('gal_deep = ',gal_deep)
    deep_flux = gal_deep.flux
    deep_hlr = gal_deep.calculateHLR()
    print('flux = ',deep_flux)
    print('hlr = ',deep_hlr)

    # Deep galaxy is fainter and smaller.
    np.testing.assert_almost_equal(deep_flux / shallow_flux, 10.**(-0.6))
    np.testing.assert_almost_equal(deep_hlr / shallow_hlr, 0.6)

    # With samples other than 23.5, it raises a warning and doesn't do any scaling.
    cat.use_sample = '25.2'
    with assert_warns(galsim.GalSimWarning):
        gal_not_deep = cat.makeGalaxy(index=17, gal_type='real', deep=True, noise_pad_size=0)
    assert gal_not_deep.flux == shallow_flux
    assert gal_not_deep.calculateHLR() == shallow_hlr

    cat.use_sample = 'user_defined'
    with assert_warns(galsim.GalSimWarning):
        gal_not_deep = cat.makeGalaxy(index=17, gal_type='real', deep=True, noise_pad_size=0)
    assert gal_not_deep.flux == shallow_flux
    assert gal_not_deep.calculateHLR() == shallow_hlr


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from copy import deepcopy
import numpy as np

import galsim
from galsim_test_helpers import *


# Save images used in regression testing for manual inspection?
save_profiles = False

# set up any necessary info for tests
# Note that changes here should match changes to test image files
image_dir = './inclined_exponential_images'

# Values here are strings, so the filenames will be sure to work (without truncating zeros)

# Tuples are in the order: (flux, n, inclination angle (radians), scale radius, scale height,
#                           truncation factor, position angle)

# Parameter sets valid for the inclined exponential case
inclined_exponential_test_parameters = (
    ("1.0", "1.0", "0.0" , "3.0", "0.3", "0.0", "0.0"),
    ("10.0", "1.0", "1.3" , "3.0", "0.5", "0.0", "0.0"),
    ("0.1", "1.0", "0.2" , "3.0", "0.5", "0.0", "0.0"),
    ("1.0", "1.0", "0.01", "3.0", "0.5", "0.0", "0.0"),
    ("2.1", "1.0", "1.57", "2.5", "0.2", "0.0", "7.4"),
    ("1.0", "1.0", "0.1" , "2.0", "1.0", "0.0", "-0.2"),
    ("1.0", "1.0", "0.78", "2.0", "0.5", "0.0", "-0.2"),)

# Parameter sets valid only for the inclined Sersic case
inclined_sersic_test_parameters = (
    ("2.0", "1.5", "0.0" , "3.0", "0.3", "0.0", "0.0"),
    ("1.5", "2.0", "0.0" , "3.0", "0.3", "0.0", "0.0"),
    ("1.1", "0.5", "0.01" , "3.0", "0.3", "0.0", "0.0"),
    ("10.0", "1.0", "0.1" , "1.9", "0.3", "4.5", "7.3"),
    ("1.0e6", "1.5", "0.2" , "2.1", "0.2", "3.9", "-0.9"),
    ("1.0e-6", "2.0", "0.3" , "3.4", "0.1", "5.0", "0.6"),
    ("2.3e4", "0.5", "1.57" , "1.8", "0.5", "2.5", "0.3"),)

# Parameter sets used for regression tests of Sersic profiles
inclined_sersic_regression_test_parameters = (
    ("1.0", "0.5", "0.7", "4.0", "0.2", "0.0", "0.5"),
    ("1.0", "0.8", "1.4", "2.5", "0.1", "4.0", "1.1"),
    ("1.0", "1.2", "2.0", "1.5", "0.5", "6.5", "5.5"),
    ("1.0", "1.5", "0.3", "3.0", "0.3", "0.0", "0.0"),)

image_nx = 64
image_ny = 64

def get_prof(mode, *args, **kwargs):
    """Function to get either InclinedExponential or InclinedSersic (with n=1, trunc=0)
       depending on mode
    """
    new_kwargs = deepcopy(kwargs)
    if len(args) > 0:
        new_kwargs["inclination"] = args[0]
    if len(args) > 1:
        new_kwargs["scale_radius"] = args[1]
    if len(args) > 2:
        new_kwargs["scale_height"] = args[2]

    if mode == "InclinedSersic":

        if not "trunc" in new_kwargs:
            new_kwargs["trunc"] = 0.
        if not "n" in new_kwargs:
            new_kwargs["n"] = 1.

        prof = galsim.InclinedSersic(**new_kwargs)
    else:
        if "trunc" in new_kwargs:
            del new_kwargs["trunc"]
        if "n" in new_kwargs:
            del new_kwargs["n"]
        if "flux_untruncated" in new_kwargs:
            del new_kwargs["flux_untruncated"]
        prof = galsim.InclinedExponential(**new_kwargs)

    return prof


@timer
def test_regression():
    """Test that the inclined exponential profile matches the results from Lance Miller's code.
       Reference images are provided in the ./inclined_exponential_images directory, as well as
       the code ('hankelcode.c') used to generate them."""

    for mode in ("InclinedExponential", "InclinedSersic"):

        for (flux, _sersic_n, inc_angle, scale_radius, scale_height,
             _trunc_factor, pos_angle) in inclined_exponential_test_parameters:

            image_filename = "galaxy_" + inc_angle + "_" + scale_radius + "_" + scale_height + "_" + pos_angle + ".fits"
            print("Comparing " + mode + " against " + image_filename + "...")

            # Get float values for the details
            flux = float(flux)
            inc_angle = float(inc_angle)
            scale_radius = float(scale_radius)
            scale_height = float(scale_height)
            pos_angle = float(pos_angle)

            image = galsim.fits.read(image_filename, image_dir)
            image *= flux / image.array.sum()
            nx, ny = np.shape(image.array)

            # Now make a test image
            test_profile = get_prof(mode, inc_angle * galsim.radians, scale_radius,
                                    scale_height, flux=flux)

            gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
            test2 = get_prof(mode, inc_angle * galsim.radians, scale_radius, scale_height,
                             flux=flux, gsparams=gsp)
            assert test2 != test_profile
            assert test2 == test_profile.withGSParams(gsp)
            assert test2 == test_profile.withGSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)

            check_basic(test_profile, mode)

            # Rotate it by the position angle
            test_profile = test_profile.rotate(pos_angle * galsim.radians)

            # Draw it onto an image
            test_image = galsim.Image(np.zeros_like(image.array), scale=1.0)
            test_profile.drawImage(test_image, offset=(0.5, 0.5)) # Offset to match Lance's

            # Save for manual inspection if desired
            if save_profiles:
                test_image_filename = image_filename.replace(".fits", "_" + mode + ".fits")
                test_image.write(test_image_filename, image_dir, clobber=True)

            # Compare to the example - Due to the different fourier transforms used, some offset is
            # expected, so we just compare in the core to two decimal places

            image_core = image.array[ nx//2-2 : nx//2+3 , ny//2-2 : ny//2+3 ]
            test_image_core = test_image.array[ nx//2-2 : nx//2+3 , ny//2-2 : ny//2+3 ]

            # Be a bit more lenient in the edge-on case, since it has greater errors in the FFT
            if np.cos(inc_angle) < 0.01:
                rtol = 5e-2
                atol = 5e-4
            else:
                rtol = 1e-2
                atol = 1e-4

            np.testing.assert_allclose(
                    test_image_core, image_core,
                    rtol=rtol, atol=atol * flux,
                    err_msg="Error in comparison of " + mode + " profile to " + image_filename,
                    verbose=True)

    # Now do Sersic-only tests
    for (flux, sersic_n, inc_angle, scale_radius, scale_height,
         trunc_factor, pos_angle) in inclined_sersic_regression_test_parameters:

        image_filename = ("galaxy_" + sersic_n + "_" + inc_angle + "_" + scale_radius +
                          "_" + scale_height + "_" + trunc_factor + "_" + pos_angle + ".fits")
        print("Comparing " + mode + " against " + image_filename + "...")

        # Get float values for the details
        flux = float(flux)
        sersic_n = float(sersic_n)
        inc_angle = float(inc_angle)
        scale_radius = float(scale_radius)
        scale_height = float(scale_height)
        trunc_factor = float(trunc_factor)
        pos_angle = float(pos_angle)

        image = galsim.fits.read(image_filename, image_dir)
        image *= flux / image.array.sum()
        nx, ny = np.shape(image.array)

        # Now make a test image
        test_profile = get_prof("InclinedSersic", n=sersic_n, scale_radius=scale_radius, scale_height=scale_height,
                                inclination=inc_angle * galsim.radians, trunc=trunc_factor * scale_radius, flux=flux)
        check_basic(test_profile, mode)

        # Rotate it by the position angle
        test_profile = test_profile.rotate(pos_angle * galsim.radians)

        # Draw it onto an image
        test_image = galsim.Image(nx, ny, scale=1.0)
        test_profile.drawImage(test_image, offset=(0.5, 0.5)) # Offset to match Lance's

        # Save if desired
        if save_profiles:
            test_image_filename = image_filename.replace(".fits", "_" + mode + ".fits")
            test_image.write(test_image_filename, image_dir, clobber=True)

        # Compare to the example - Due to the different fourier transforms used, some offset is
        # expected, so we just compare in the core to two decimal places

        image_core = image.array[nx // 2 - 2:nx // 2 + 3, ny // 2 - 2:ny // 2 + 3]
        test_image_core = test_image.array[nx // 2 - 2:nx // 2 + 3, ny // 2 - 2:ny // 2 + 3]

        np.testing.assert_allclose(
                test_image_core, image_core,
                rtol=1e-2, atol=1e-4 * flux,
                err_msg="Error in comparison of " + mode + " profile to " + image_filename,
                verbose=True)


@timer
def test_exponential():
    """ Test that InclinedExponential looks identical to an exponential when inclination is zero.
    """

    scale_radius = 3.0
    hlr = 1.7
    mode = "InclinedExponential"

    # Construct from scale_radius
    exp_profile = galsim.Exponential(scale_radius=scale_radius)
    inc_profile = get_prof(mode, 0 * galsim.radians, scale_radius=scale_radius,
                           scale_height=scale_radius / 10.)
    np.testing.assert_almost_equal(inc_profile.scale_radius, exp_profile.scale_radius)
    np.testing.assert_almost_equal(inc_profile.disk_half_light_radius,
                                   exp_profile.half_light_radius)

    # Construct from half_light_radius
    exp_profile = galsim.Exponential(half_light_radius=hlr)
    inc_profile = get_prof(mode, 0 * galsim.radians, half_light_radius=hlr,
                           scale_height=scale_radius / 10.)
    np.testing.assert_almost_equal(inc_profile.scale_radius, exp_profile.scale_radius)
    np.testing.assert_almost_equal(inc_profile.disk_half_light_radius,
                                   exp_profile.half_light_radius)

    exp_image = galsim.Image(image_nx, image_ny, scale=1.0)
    exp_profile.drawImage(exp_image)
    inc_image = galsim.Image(image_nx, image_ny, scale=1.0)
    inc_profile.drawImage(inc_image)

    # Check that they're the same
    np.testing.assert_array_almost_equal(inc_image.array, exp_image.array, decimal=4)

    # The face-on version should get the max_sb value exactly right
    np.testing.assert_array_almost_equal(inc_profile.max_sb, exp_profile.max_sb)

    check_basic(inc_profile, "Face-on " + mode)


@timer
def test_sersic():
    """ Test that InclinedSersic looks identical to a Sersic when inclination is zero.
    """

    ns = (1.1, 1.1, 2.5, 2.5)
    truncs = (0, 13.5, 0, 18.0)
    scale_radius = 3.0
    hlr = 1.7
    mode = "InclinedSersic"

    for n, trunc in zip(ns, truncs):

        # Construct from scale_radius
        sersic_profile = galsim.Sersic(n=n, scale_radius=scale_radius, trunc=trunc)
        inc_profile = get_prof(mode, n=n, trunc=trunc, inclination=0 * galsim.radians,
                               scale_radius=scale_radius,
                               scale_height=hlr / 10.)
        np.testing.assert_almost_equal(inc_profile.scale_radius, sersic_profile.scale_radius)
        np.testing.assert_almost_equal(inc_profile.disk_half_light_radius,
                                       sersic_profile.half_light_radius)

        # Construct from half-light radius
        sersic_profile = galsim.Sersic(n=n, half_light_radius=hlr, trunc=trunc)
        inc_profile = get_prof(mode, n=n, trunc=trunc, inclination=0 * galsim.radians,
                               half_light_radius=hlr,
                               scale_height=hlr / 10.)
        np.testing.assert_almost_equal(inc_profile.scale_radius, sersic_profile.scale_radius)
        np.testing.assert_almost_equal(inc_profile.disk_half_light_radius,
                                       sersic_profile.half_light_radius)

        sersic_image = galsim.Image(image_nx, image_ny, scale=1.0)
        sersic_profile.drawImage(sersic_image)
        inc_image = galsim.Image(image_nx, image_ny, scale=1.0)
        inc_profile.drawImage(inc_image)

        if save_profiles:
            sersic_image.write("test_sersic.fits", image_dir, clobber=True)
            inc_image.write("test_inclined_sersic.fits", image_dir, clobber=True)

        # Check that they're the same. Note that since the inclined Sersic profile isn't
        # Real-space analytic and has hard edges in the truncated case,
        # we have to be a bit lax on rtol and atol
        if trunc != 0:
            rtol = 5e-3
            atol = 5e-5
        else:
            rtol = 1e-3
            atol = 1e-5

        np.testing.assert_allclose(inc_image.array, sersic_image.array, rtol=rtol, atol=atol)

        # The face-on version should get the max_sb value exactly right
        np.testing.assert_almost_equal(inc_profile.max_sb, sersic_profile.max_sb)

        check_basic(inc_profile, "Face-on " + mode)


@timer
def test_edge_on():
    """ Test that an edge-on profile looks similar to an almost-edge-on profile, and doesn't crash.
    """
    from scipy.special import gamma

    scale_radius = 3.0
    sersic_n = 2.0

    inclinations = (np.arccos(0.01), 2 * np.pi - np.arccos(0.01), np.pi / 2.)

    for mode in ("InclinedExponential", "InclinedSersic"):

        if mode == "InclinedExponential":
            n = 1.0
            comp_prof = galsim.Exponential(scale_radius=scale_radius)
        else:
            n = sersic_n
            comp_prof = galsim.Sersic(n=n, scale_radius=scale_radius)

        images = []

        for inclination in inclinations:
            # Set up the profile
            prof = get_prof(mode, inclination * galsim.radians, scale_radius=scale_radius,
                            scale_h_over_r=0.1, n=n)

            check_basic(prof, "Edge-on " + mode)

            # Draw an image of it
            image = galsim.Image(image_nx, image_ny, scale=1.0)
            prof.drawImage(image)

            # Add it to the list of images
            images.append(image.array)

        # Check they're all almost the same
        np.testing.assert_array_almost_equal(images[1], images[0], decimal=2)
        np.testing.assert_array_almost_equal(images[1], images[2], decimal=2)

        # Also the edge-on version should get the max_sb value exactly right
        np.testing.assert_allclose(prof.max_sb, comp_prof.max_sb * 10. * n / gamma(n))
        prof.drawImage(image, method='sb', use_true_center=False)
        print('max pixel: ', image.array.max(), ' cf.', prof.max_sb)
        np.testing.assert_allclose(image.array.max(), prof.max_sb, rtol=0.01)


@timer
def test_sanity():
    """ Performs various sanity checks on a set of InclinedExponential and InclinedSersic profiles. """

    def run_sanity_checks(mode, flux, inc_angle, scale_radius, scale_height, pos_angle,
                          n=1., trunc=0.):
        # Get float values for the details
        print(flux, inc_angle, scale_radius, scale_height, pos_angle, n, trunc)

        # Now make a test image
        test_profile = get_prof(mode, inc_angle * galsim.radians, scale_radius,
                                scale_height, flux=flux, n=n, trunc=trunc)

        check_basic(test_profile, mode)

        # Check accessing construction args
        np.testing.assert_equal(test_profile.inclination, inc_angle * galsim.radians)
        np.testing.assert_equal(test_profile.scale_radius, scale_radius)
        np.testing.assert_equal(test_profile.scale_height, scale_height)
        np.testing.assert_equal(test_profile.flux, flux)

        # Check that h/r is properly given by the method and property for it
        np.testing.assert_allclose(test_profile.scale_height / test_profile.scale_radius,
                                       test_profile.scale_h_over_r, rtol=1e-4)

        # Rotate it by the position angle
        test_profile = test_profile.rotate(pos_angle * galsim.radians)

        # Check that the k value for (0,0) is the flux
        np.testing.assert_allclose(test_profile.kValue(kx=0., ky=0.), flux, rtol=1e-4)

        # Check that the drawn flux for a large image is indeed the flux
        test_image = galsim.Image(int(5 * n ** 2) * image_nx, int(5 * n ** 2) * image_ny, scale=1.0)
        test_profile.drawImage(test_image)
        test_flux = test_image.array.sum()
        # Be a bit more lenient here for sersic profiles than exponentials
        if n == 1:
            rtol = 1e-4
        else:
            rtol = 1e-2
        np.testing.assert_allclose(test_flux, flux, rtol=rtol)

        # Check that the centroid is (0,0)
        centroid = test_profile.centroid
        np.testing.assert_equal(centroid.x, 0.)
        np.testing.assert_equal(centroid.y, 0.)

        # Check max_sb - just ensure it's not under by more than the empirical limit for this
        # approximation
        test_profile.drawImage(test_image, use_true_center=False)
        print('max pixel: ', test_image.array.max(), ' cf.', test_profile.max_sb)

        np.testing.assert_array_less(test_image.array.max(), test_profile.max_sb*1.56)

    # Run tests applicable to both profiles
    for mode in ("InclinedExponential", "InclinedSersic"):

        print('flux, inc_angle, scale_radius, scale_height, pos_angle, n, trunc')
        for (flux, _sersic_n, inc_angle, scale_radius, scale_height,
             _trunc_factor, pos_angle) in inclined_exponential_test_parameters:

            flux = float(flux)
            inc_angle = float(inc_angle)
            scale_radius = float(scale_radius)
            scale_height = float(scale_height)
            pos_angle = float(pos_angle)
            run_sanity_checks(mode, flux, inc_angle, scale_radius, scale_height, pos_angle)

    # Run tests for InclinedSersic only
    for (flux, sersic_n, inc_angle, scale_radius, scale_height,
         trunc_factor, pos_angle) in (inclined_sersic_test_parameters +
                                      inclined_sersic_regression_test_parameters):
        flux = float(flux)
        inc_angle = float(inc_angle)
        scale_radius = float(scale_radius)
        scale_height = float(scale_height)
        pos_angle = float(pos_angle)
        n = float(sersic_n)
        trunc = float(trunc_factor) * float(scale_radius)

        run_sanity_checks("InclinedSersic", flux, inc_angle, scale_radius, scale_height, pos_angle,
                          n=n, trunc=trunc)

        # Run specific tests for InclinedSersic

        # Check that flux_untruncated behaves as expected
        prof1a = get_prof("InclinedSersic",n=n,flux=flux,inclination=inc_angle*galsim.radians,
                         scale_radius=scale_radius,scale_height=scale_height,trunc=trunc)
        prof1b = get_prof("InclinedSersic",n=n,flux=flux,inclination=inc_angle*galsim.radians,
                         scale_radius=scale_radius,scale_height=scale_height,trunc=trunc,
                         flux_untruncated=False)
        prof2 = get_prof("InclinedSersic",n=n,flux=flux,inclination=inc_angle*galsim.radians,
                         scale_radius=scale_radius,scale_height=scale_height,trunc=trunc,
                         flux_untruncated=True)

        np.testing.assert_almost_equal(prof1a.flux, prof1b.flux, 9)
        print(flux, trunc_factor, trunc)
        print("   ", prof1a.flux, prof1b.flux, prof2.flux)
        if trunc > 0:
            assert(prof1a.flux > prof2.flux)


@timer
def test_k_limits(run_slow):
    """ Check that the maxk and stepk give reasonable results for all profiles. """

    test_params = inclined_sersic_regression_test_parameters
    if run_slow:
        test_params += inclined_exponential_test_parameters
        test_params += inclined_sersic_test_parameters

    for mode in ("InclinedExponential", "InclinedSersic"):

        for (_, _, inc_angle, scale_radius, scale_height, _, _) in test_params:

            # Get float values for the details
            inc_angle = float(inc_angle)
            scale_radius = float(scale_radius)
            scale_height = float(scale_height)

            gsparams = galsim.GSParams()

            # Now make a test image
            test_profile = get_prof(mode, inc_angle * galsim.radians, scale_radius, scale_height)

            # Check that the k value at maxk is below maxk_threshold in both the x and y dimensions
            kx = test_profile.maxk
            ky = test_profile.maxk

            kx_value = test_profile.kValue(kx=kx, ky=0.)
            np.testing.assert_(np.abs(kx_value) < gsparams.maxk_threshold,
                               msg="kx_value is not below maxk_threshold: " + str(kx_value) + " >= "
                                + str(gsparams.maxk_threshold))

            ky_value = test_profile.kValue(kx=0., ky=ky)
            np.testing.assert_(np.abs(ky_value) < gsparams.maxk_threshold,
                               msg="ky_value is not below maxk_threshold: " + str(ky_value) + " >= "
                                + str(gsparams.maxk_threshold))

            # Check that less than folding_threshold fraction of light falls outside r = pi/stepk
            rmax = np.pi / test_profile.stepk
            pixel_scale = 0.1
            test_image = galsim.Image(int(2*rmax/pixel_scale), int(2*rmax/pixel_scale),
                                      scale=pixel_scale)
            test_profile.drawImage(test_image)

            # Get an array of indices within the limits
            image_shape = np.shape(test_image.array)
            x, y = np.indices(image_shape, dtype=float)

            image_center = test_image.center
            x -= image_center.x
            y -= image_center.y

            # Include all pixels that are at least partially within distance r of the centre
            r = pixel_scale * np.sqrt(np.square(x) + np.square(y))
            good = r < rmax + np.sqrt(2.)*pixel_scale

            # Get flux within the limits
            contained_flux = np.ravel(test_image.array)[np.ravel(good)].sum()

            # Check that we're not missing too much flux
            total_flux = np.sum(test_image.array)
            assert (total_flux-contained_flux)/total_flux <= gsparams.folding_threshold

            # Check that stepk and maxk scale correctly with size
            test_profile2 = get_prof(mode, inc_angle * galsim.radians,
                                     5*scale_radius, 5*scale_height)
            np.testing.assert_almost_equal(test_profile2.maxk, test_profile.maxk/5)
            np.testing.assert_almost_equal(test_profile2.stepk, test_profile.stepk/5)

@timer
def test_eq_ne():
    """ Check that equality/inequality works as expected."""

    gsp = galsim.GSParams(folding_threshold=1.1e-3)

    diff_gals = []

    for mode in ("InclinedExponential", "InclinedSersic"):

        # First test that some different initializations that should be equivalent:
        same_gals = [get_prof(mode, 0.1 * galsim.radians, 3.0),
                get_prof(mode, 0.1 * galsim.radians, 3.0, 0.3), # default h/r = 0.1
                get_prof(mode, 0.1 * galsim.radians, 3.0, scale_height=0.3),
                get_prof(mode, 0.1 * galsim.radians, 3.0, scale_h_over_r=0.1),
                get_prof(mode, 0.1 * galsim.radians, 3.0, flux=1.0), # default flux=1
                get_prof(mode, -0.1 * galsim.radians, 3.0), # negative i is equivalent
                get_prof(mode, (np.pi - 0.1) * galsim.radians, 3.0), # also pi-theta
                get_prof(mode, 18. / np.pi * galsim.degrees, 3.0),
                get_prof(mode, inclination=0.1 * galsim.radians, scale_radius=3.0,
                         scale_height=0.3, flux=1.0),
                get_prof(mode, flux=1.0, scale_radius=3.0,
                         scale_height=0.3, inclination=0.1 * galsim.radians),
                get_prof(mode, flux=1.0, half_light_radius=3.0 * galsim.Exponential._hlr_factor,
                         scale_height=0.3, inclination=0.1 * galsim.radians),
                get_prof(mode, flux=1.0, half_light_radius=3.0 * galsim.Exponential._hlr_factor,
                         scale_h_over_r=0.1, inclination=0.1 * galsim.radians)]

        for gal in same_gals[1:]:
            print(gal)
            gsobject_compare(gal, same_gals[0])

        # Set up list of galaxies we expect to all be different
        diff_gals += [get_prof(mode, 0.1 * galsim.radians, 3.0, 0.3),
                get_prof(mode, 0.1 * galsim.degrees, 3.0, 0.3),
                get_prof(mode, 0.1 * galsim.degrees, 3.0, scale_h_over_r=0.2),
                get_prof(mode, 0.1 * galsim.radians, 3.0, 3.0),
                get_prof(mode, 0.2 * galsim.radians, 3.0, 0.3),
                get_prof(mode, 0.1 * galsim.radians, 3.1, 0.3),
                get_prof(mode, 0.1 * galsim.radians, 3.1),
                get_prof(mode, 0.1 * galsim.radians, 3.0, 0.3, flux=0.5),
                get_prof(mode, 0.1 * galsim.radians, 3.0, 0.3, gsparams=gsp)]

    # Add some more Sersic profiles to the diff_gals list
    diff_gals += [get_prof("InclinedSersic", 0.1 * galsim.radians, 3.0, 0.3, n=1.1),
                  get_prof("InclinedSersic", 0.1 * galsim.radians, 3.0, 0.3, trunc=4.5),
                  galsim.InclinedSersic(n=1.0, inclination=0.1 * galsim.radians, half_light_radius=3.0,
                                        scale_height=0.3),
                  galsim.InclinedSersic(n=1.0, inclination=0.1 * galsim.radians, scale_radius=3.0,
                                        scale_h_over_r=0.3)]

    check_all_diff(diff_gals)


@timer
def test_pickle():
    """ Check that we can pickle it. """

    for mode in ("InclinedExponential", "InclinedSersic"):

        prof = get_prof(mode, trunc=4.5, inclination=0.1 * galsim.radians, scale_radius=3.0,
                                         scale_height=0.3)
        check_pickle(prof)

        check_pickle(get_prof(mode, trunc=4.5, inclination=0.1 * galsim.radians, scale_radius=3.0))
        check_pickle(get_prof(mode, trunc=4.5, inclination=0.1 * galsim.radians, scale_radius=3.0,
                                             scale_h_over_r=0.2))
        check_pickle(get_prof(mode, trunc=4.5, inclination=0.1 * galsim.radians, scale_radius=3.0,
                                             scale_height=0.3, flux=10.0))
        check_pickle(get_prof(mode, trunc=4.5, inclination=0.1 * galsim.radians, scale_radius=3.0,
                                             scale_height=0.3,
                                             gsparams=galsim.GSParams(folding_threshold=1.1e-3)))
        check_pickle(get_prof(mode, trunc=4.5, inclination=0.1 * galsim.radians, scale_radius=3.0,
                                             scale_height=0.3, flux=10.0,
                                             gsparams=galsim.GSParams(folding_threshold=1.1e-3)))


@timer
def test_exceptions():
    """ Tests to make sure that proper exceptions are thrown when expected. """

    for mode in ("InclinedExponential", "InclinedSersic"):

        # Need at least one radius specification
        with assert_raises(TypeError):
            get_prof(mode, inclination = 0.*galsim.degrees)

        # Can't have two radius specifications
        with assert_raises(TypeError):
            get_prof(mode, inclination = 0.*galsim.degrees,
                     scale_radius = 1., half_light_radius = 1.)

        # Radius specification must be > 0
        with assert_raises(ValueError):
            get_prof(mode, inclination = 0.*galsim.degrees, scale_radius = -1.)
        with assert_raises(ValueError):
            get_prof(mode, inclination = 0.*galsim.degrees, half_light_radius = -1.)

        # Can't have both height specifications
        with assert_raises(TypeError):
            get_prof(mode, inclination = 0.*galsim.degrees,
                     scale_radius = 1., scale_height = 0.2, scale_h_over_r = 0.1)

        # Radius specification must be > 0
        with assert_raises(ValueError):
            get_prof(mode, inclination = 0.*galsim.degrees, scale_radius = 1., scale_height = -0.2)
        with assert_raises(ValueError):
            get_prof(mode, inclination = 0.*galsim.degrees,
                     scale_radius = 1., scale_h_over_r = -0.1)

        # Enforce inclination is an angle type
        with assert_raises(TypeError):
            get_prof(mode, inclination = 0., scale_radius = 1.)

    # Can't have negative truncation for InclinedSersic
    with assert_raises(ValueError):
        get_prof("InclinedSersic", inclination = 0.*galsim.degrees,
                 scale_radius = 1., trunc = -4.5)

    # trunc can't be too small in InclinedSersic
    with assert_raises(ValueError):
        get_prof("InclinedSersic", inclination = 0.*galsim.degrees,
                 half_light_radius = 1., trunc = 1.4)


@timer
def test_value_retrieval():
    """ Tests to make sure that if a parameter is passed to a profile, we get back the same
        value from it. Only parameters not tested by the pickling are tested here.
    """

    for mode in ("InclinedExponential", "InclinedSersic"):

        half_light_radius = 1.6342
        scale_h_over_r = 0.2435

        prof = get_prof(mode, inclination=0.*galsim.degrees, half_light_radius=half_light_radius,
                        scale_h_over_r=scale_h_over_r)

        np.testing.assert_almost_equal(half_light_radius, prof.disk_half_light_radius, 9)
        np.testing.assert_almost_equal(scale_h_over_r, prof.scale_h_over_r, 9)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os
import sys

import galsim
from galsim_test_helpers import *

path, filename = os.path.split(__file__)
imgdir = os.path.join(path, "SBProfile_comparison_images") # Directory containing the reference
                                                           # images.

@timer
def test_kolmogorov():
    """Test the generation of a specific Kolmogorov profile against a known result.
    """
    import math
    dx = 0.2
    test_flux = 1.8
    # This savedImg was created from the SBKolmogorov implementation in
    # commit c8efd74d1930157b1b1ffc0bfcfb5e1bf6fe3201
    # It would be nice to get an independent calculation here...
    #mySBP = galsim.SBKolmogorov(lam_over_r0=1.5, flux=test_flux)
    #savedImg = galsim.ImageF(128,128)
    #mySBP.drawImage(image=savedImg, dx=dx, method="sb")
    #savedImg.write(os.path.join(imgdir, "kolmogorov.fits"))
    savedImg = galsim.fits.read(os.path.join(imgdir, "kolmogorov.fits"))
    myImg = galsim.ImageF(savedImg.bounds, scale=dx)
    myImg.setCenter(0,0)

    kolm = galsim.Kolmogorov(lam_over_r0=1.5, flux=test_flux)
    kolm.drawImage(myImg, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Kolmogorov disagrees with expected result")

    # Check with default_params
    kolm = galsim.Kolmogorov(lam_over_r0=1.5, flux=test_flux, gsparams=default_params)
    kolm.drawImage(myImg, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Kolmogorov with default_params disagrees with expected result")
    kolm = galsim.Kolmogorov(lam_over_r0=1.5, flux=test_flux, gsparams=galsim.GSParams())
    kolm.drawImage(myImg, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Kolmogorov with GSParams() disagrees with expected result")

    gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
    kolm2 = galsim.Kolmogorov(lam_over_r0=1.5, flux=test_flux, gsparams=gsp)
    assert kolm2 != kolm
    assert kolm2 == kolm.withGSParams(gsp)
    assert kolm2 == kolm.withGSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)

    check_basic(kolm, "Kolmogorov")

    # Test photon shooting.
    do_shoot(kolm,myImg,"Kolmogorov")

    # Test kvalues
    do_kvalue(kolm,myImg, "Kolmogorov")

    # Check picklability
    check_pickle(kolm, lambda x: x.drawImage(method='no_pixel'))
    check_pickle(kolm)

    # Test initialization separately with lam and r0, in various units.  Since the above profiles
    # have lam/r0 = 3./2. in arbitrary units, we will tell it that lam=3.e9 nm and r0=2.0 m,
    # and use `scale_unit` of galsim.radians.  This is rather silly, but it should work.
    kolm = galsim.Kolmogorov(lam_over_r0=1.5, flux=test_flux)
    kolm2 = galsim.Kolmogorov(lam=3.e9, r0=2.0, scale_unit=galsim.radians, flux=test_flux)
    gsobject_compare(kolm,kolm2)
    # For lam/r0 = 1.5 arcsec, and r0 = 0.2, lam = (1.5/3600/180*pi) * 0.2 * 1.e9
    lam = 1.5 * 0.2 / 3600. / 180. * math.pi * 1.e9
    print('lam = ',lam)
    kolm3 = galsim.Kolmogorov(lam=lam, r0=0.2, scale_unit='arcsec', flux=test_flux)
    gsobject_compare(kolm,kolm3)
    # arcsec is the default scale_unit, so can leave this off.
    kolm4 = galsim.Kolmogorov(lam=lam, r0=0.2, flux=test_flux)
    gsobject_compare(kolm,kolm4)
    # Test using r0_500 instead
    r0_500 = 0.2 * (lam/500)**-1.2
    kolm5 = galsim.Kolmogorov(lam=lam, r0_500=r0_500, flux=test_flux)
    gsobject_compare(kolm,kolm5)

    # Should raise an exception if >= 2 radius specifications are provided and/or lam and r0 are not
    # paired together.
    assert_raises(TypeError, galsim.Kolmogorov,
                  lam_over_r0=3, fwhm=2, half_light_radius=1, lam=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, fwhm=2, half_light_radius=1, lam=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, half_light_radius=1, lam=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, fwhm=2, lam=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, fwhm=2, half_light_radius=1)
    assert_raises(TypeError, galsim.Kolmogorov, half_light_radius=1, lam=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, fwhm=2, lam=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, fwhm=2, half_light_radius=1)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, lam=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, half_light_radius=1)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, fwhm=2)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, lam=3)
    assert_raises(TypeError, galsim.Kolmogorov, lam_over_r0=3, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, fwhm=2, lam=3)
    assert_raises(TypeError, galsim.Kolmogorov, fwhm=2, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, half_light_radius=1, lam=3)
    assert_raises(TypeError, galsim.Kolmogorov, half_light_radius=1, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov, lam=3)
    assert_raises(TypeError, galsim.Kolmogorov, r0=1)
    assert_raises(TypeError, galsim.Kolmogorov)

@timer
def test_kolmogorov_properties():
    """Test some basic properties of the Kolmogorov profile.
    """
    test_flux = 1.8
    lor = 1.5
    psf = galsim.Kolmogorov(lam_over_r0=lor, flux=test_flux)
    # Check that we are centered on (0, 0)
    cen = galsim.PositionD(0, 0)
    np.testing.assert_equal(psf.centroid, cen)
    # Check Fourier properties
    np.testing.assert_almost_equal(psf.maxk, 8.644067599028375, 5)
    np.testing.assert_almost_equal(psf.stepk, 0.379103528681262, 5)
    np.testing.assert_almost_equal(psf.kValue(cen), test_flux+0j)
    np.testing.assert_almost_equal(psf.lam_over_r0, lor)
    np.testing.assert_almost_equal(psf.half_light_radius, lor * 0.5548101137)
    np.testing.assert_almost_equal(psf.fwhm, lor * 0.9758634299)
    np.testing.assert_almost_equal(psf.xValue(cen), 0.6283185307179587)
    np.testing.assert_almost_equal(psf.kValue(cen), (1+0j) * test_flux)
    np.testing.assert_almost_equal(psf.flux, test_flux)
    np.testing.assert_almost_equal(psf.xValue(cen), psf.max_sb)

    # Check input flux vs output flux
    lors = [1, 0.5, 2, 5]
    for lor in lors:
        psf = galsim.Kolmogorov(lam_over_r0=lor, flux=test_flux)
        out_flux = psf.flux
        np.testing.assert_almost_equal(out_flux, test_flux,
                                       err_msg="Flux of Kolmogorov is incorrect.")

        # Also check the realized flux in a drawn image
        dx = lor / 10.
        img = galsim.ImageF(256,256, scale=dx)
        psf.drawImage(image=img)
        out_flux = img.array.sum(dtype=float)
        np.testing.assert_almost_equal(out_flux, test_flux, 3,
                                       err_msg="Flux of Kolmogorov (image array) is incorrect.")

    # Check that stepk and maxk scale correctly with radius
    psf2 = galsim.Kolmogorov(lam_over_r0=5*lor, flux=test_flux)
    np.testing.assert_almost_equal(psf2.maxk, psf.maxk/5)
    np.testing.assert_almost_equal(psf2.stepk, psf.stepk/5)


@timer
def test_kolmogorov_radii():
    """Test initialization of Kolmogorov with different types of radius specification.
    """
    import math
    test_hlr = 1.9
    test_fwhm = 1.8

    # Test constructor using lambda/r0
    lors = [1, 0.5, 2, 5]
    for lor in lors:
        print('lor = ',lor)
        test_gal = galsim.Kolmogorov(flux=1., lam_over_r0=lor)

        np.testing.assert_almost_equal(
                lor, test_gal.lam_over_r0, decimal=9,
                err_msg="Error in Kolmogorov, lor != lam_over_r0")

        # test that fwhm provides correct FWHM
        got_fwhm = test_gal.fwhm
        print('got_fwhm = ',got_fwhm)
        test_fwhm_ratio = (test_gal.xValue(galsim.PositionD(.5 * got_fwhm, 0.)) /
                        test_gal.xValue(galsim.PositionD(0., 0.)))
        print('fwhm ratio = ', test_fwhm_ratio)
        np.testing.assert_almost_equal(
                test_fwhm_ratio, 0.5, decimal=4,
                err_msg="Error in FWHM for Kolmogorov initialized with half-light radius")

        # then test that image indeed has the correct HLR properties when radially integrated
        got_hlr = test_gal.half_light_radius
        print('got_hlr = ',got_hlr)
        hlr_sum = radial_integrate(test_gal, 0., got_hlr)
        print('hlr_sum = ',hlr_sum)
        np.testing.assert_almost_equal(
                hlr_sum, 0.5, decimal=3,
                err_msg="Error in half light radius for Kolmogorov initialized with lam_over_r0.")

    # Test constructor using half-light-radius:
    test_gal = galsim.Kolmogorov(flux=1., half_light_radius=test_hlr)
    hlr_sum = radial_integrate(test_gal, 0., test_hlr)
    print('hlr_sum = ',hlr_sum)
    np.testing.assert_almost_equal(
            hlr_sum, 0.5, decimal=3,
            err_msg="Error in Kolmogorov constructor with half-light radius")

    # test that fwhm provides correct FWHM
    got_fwhm = test_gal.fwhm
    print('got_fwhm = ',got_fwhm)
    test_fwhm_ratio = (test_gal.xValue(galsim.PositionD(.5 * got_fwhm, 0.)) /
                    test_gal.xValue(galsim.PositionD(0., 0.)))
    print('fwhm ratio = ', test_fwhm_ratio)
    np.testing.assert_almost_equal(
            test_fwhm_ratio, 0.5, decimal=4,
            err_msg="Error in FWHM for Kolmogorov initialized with half-light radius")

    # Test constructor using FWHM:
    test_gal = galsim.Kolmogorov(flux=1., fwhm = test_fwhm)
    center = test_gal.xValue(galsim.PositionD(0,0))
    ratio = test_gal.xValue(galsim.PositionD(test_fwhm/2.,0)) / center
    print('fwhm ratio = ',ratio)
    np.testing.assert_almost_equal(
            ratio, 0.5, decimal=4,
            err_msg="Error in Kolmogorov constructor with fwhm")

    # then test that image indeed has the correct HLR properties when radially integrated
    got_hlr = test_gal.half_light_radius
    print('got_hlr = ',got_hlr)
    hlr_sum = radial_integrate(test_gal, 0., got_hlr)
    print('hlr_sum (profile initialized with fwhm) = ',hlr_sum)
    np.testing.assert_almost_equal(
            hlr_sum, 0.5, decimal=3,
            err_msg="Error in half light radius for Kolmogorov initialized with FWHM.")

    # Check that the getters don't work after modifying the original.
    test_gal_shear = test_gal.shear(g1=0.3, g2=0.1)
    assert_raises(AttributeError, getattr, test_gal_shear, "fwhm")
    assert_raises(AttributeError, getattr, test_gal_shear, "half_light_radius")
    assert_raises(AttributeError, getattr, test_gal_shear, "lam_over_r0")


@timer
def test_kolmogorov_flux_scaling():
    """Test flux scaling for Kolmogorov.
    """
    # decimal point to go to for parameter value comparisons
    param_decimal = 12
    test_lor0 = 1.9
    test_flux = 17.9

    # init with lam_over_r0 and flux only (should be ok given last tests)
    obj = galsim.Kolmogorov(lam_over_r0=test_lor0, flux=test_flux)
    obj *= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __imul__.")
    obj = galsim.Kolmogorov(lam_over_r0=test_lor0, flux=test_flux)
    obj /= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __idiv__.")
    obj = galsim.Kolmogorov(lam_over_r0=test_lor0, flux=test_flux)
    obj2 = obj * 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (result).")
    obj = galsim.Kolmogorov(lam_over_r0=test_lor0, flux=test_flux)
    obj2 = 2. * obj
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (result).")
    obj = galsim.Kolmogorov(lam_over_r0=test_lor0, flux=test_flux)
    obj2 = obj / 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (result).")


@timer
def test_kolmogorov_folding_threshold():
    """Test Kolmogorov with low folding_threshold.
    """
    # This test reproduces a bug reported by Jim Chiang when Kolmogorov has a less than
    # default folding_threshold.  Reported in Issue #952.
    # It turned out the problem was in OneDimensionalDeviate's findExtremum going into an
    # endless loop, mostly because it didn't preserve the intended invariant that x1 < x2 < x3.
    # In this case, x1 became the largest of the three values and ended up getting larger and
    # larger indefinitely, thus resulting in an endless loop.

    # The test is really just to construct the object in finite time, but we do a few sanity
    # checks afterward for good measure.

    fwhm = 0.55344217545630736
    folding_threshold=0.0008316873626901008
    gsparams = galsim.GSParams(folding_threshold=folding_threshold)

    obj = galsim.Kolmogorov(fwhm=fwhm, gsparams=gsparams)
    print('obj = ',obj)

    assert obj.flux == 1.0
    assert obj.fwhm == fwhm
    assert obj.gsparams.folding_threshold == folding_threshold
    check_basic(obj, 'Kolmogorov with low folding_threshold')
    check_pickle(obj)


@timer
def test_kolmogorov_shoot():
    """Test Kolmogorov with photon shooting.  Particularly the flux of the final image.
    """
    rng = galsim.BaseDeviate(1234)
    obj = galsim.Kolmogorov(fwhm=1.5, flux=1.e4)
    im = galsim.Image(500,500, scale=1)
    im.setCenter(0,0)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng.duplicate())
    assert photons2 == photons, "Kolmogorov makePhot not equivalent to drawPhot"

    # Can treat the profile as a convolution of a delta function and put it in a photon_ops list.
    delta = galsim.DeltaFunction(flux=1.e4)
    psf = galsim.Kolmogorov(fwhm=1.5)
    photons3 = delta.makePhot(poisson_flux=False, rng=rng.duplicate(), photon_ops=[psf])
    assert photons3 == photons, "Using Kolmogorov in photon_ops not equivalent to drawPhot"


@timer
def test_ne():
    """Test base.py GSObjects for not-equals."""
    # Define some universal gsps
    gsp = galsim.GSParams(maxk_threshold=1.1e-3, folding_threshold=5.1e-3)

    # Kolmogorov.  Params include lam_over_r0, fwhm, half_light_radius, lam/r0, lam/r0_500, flux
    # gsparams.
    # The following should all test unequal:
    gals = [galsim.Kolmogorov(lam_over_r0=1.0),
            galsim.Kolmogorov(lam=1.0, r0=1.1),
            galsim.Kolmogorov(fwhm=1.0),
            galsim.Kolmogorov(half_light_radius=1.0),
            galsim.Kolmogorov(lam=1.0, r0=1.0),
            galsim.Kolmogorov(lam=1.0, r0=1.0, scale_unit=galsim.arcmin),
            galsim.Kolmogorov(lam=1.0, r0=1.0, scale_unit='degrees'),
            galsim.Kolmogorov(lam=1.0, r0_500=1.0),
            galsim.Kolmogorov(lam=1.0, r0=1.0, flux=1.1),
            galsim.Kolmogorov(lam=1.0, r0=1.0, flux=1.1, gsparams=gsp)]
    check_all_diff(gals)

@timer
def test_low_folding_threshold():
    """Test Kolmogorov with a very low folding_threshold.
    """
    # Jim Chiang identified a bug introduced in commit 56003a938963ba4bef875c2
    # where stepk was much too large for Kolmogorov profiles with very low folding_threshold.
    # This test checks that we have in fact fixed the bug.
    folding_threshold = 1e-6
    pixel_scale = 0.2

    gsparams = galsim.GSParams(folding_threshold=folding_threshold)

    fwhm = 0.7313  # This was the particular fwhm from Jim's test showing the failure.
    psf = galsim.Kolmogorov(fwhm=fwhm, gsparams=gsparams)
    image_size = psf.getGoodImageSize(pixel_scale)
    print('ft = 1.e-6: psf.getGoodImageSize:', image_size)
    # Note: Older versions gave 574 for this choice, which was actually too small due to
    # numerical rounding errors. Then it was wrongly zero for a while (only on main, not on
    # a release branch), and now I think this is correct.
    assert image_size == 6862

    # I think going forward, 1.e-6 is much too small a folding_threshold for the desired
    # effect Jim wants in imSim.  I think 1.e-4 is more appropriate to get a stamp that will
    # include most of the visible photons from a bright star.
    folding_threshold = 1e-4
    gsparams = galsim.GSParams(folding_threshold=folding_threshold)
    psf = galsim.Kolmogorov(fwhm=fwhm, gsparams=gsparams)
    image_size = psf.getGoodImageSize(pixel_scale)
    print('ft = 1.e-4: psf.getGoodImageSize:', image_size)
    assert image_size == 434


if __name__ == "__main__":
    runtests(__file__)
